(*
   GUI.pas - Main GUI unit

   Copyright (C) 2013 -2015 Martin Bittermann (martinbittermann@gmx.de)

   This file is part of ddrescueview.

   ddrescueview is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ddrescueview is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ddrescueview.  If not, see <http://www.gnu.org/licenses/>.
*)

unit GUI;

interface

uses
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus, ExtCtrls,
  ComCtrls, StdCtrls, Buttons, Shared, Parser, ImgBlockStore, graphtype;

type
  { TMainForm }
  TMainForm = class(TForm, IFPObserver, IAttachableObserved)
    BottomPanel: TPanel;
    CommentsMemo: TMemo;
    RSCol1Panel: TPanel;
    RSCol2Panel: TPanel;
    EditErrorCount: TEdit;
    EditInputSize: TEdit;
    EditCurrentPos: TEdit;
    EditErrorSize: TEdit;
    EditWhatever: TEdit;
    EditRescued: TEdit;
    miDevExample: TMenuItem;
    RescueStatusBox: TGroupBox;
    BtnCloseZoomBar: TSpeedButton;
    lblErrorCount: TLabel;
    lblCurrentPos: TLabel;
    lblRescued2: TLabel;
    lblWhatever: TLabel;
    lblErrorSize: TLabel;
    lblInputSize: TLabel;
    MainPanel: TPanel;
    miShowLog: TMenuItem;
    BottomHSplitter: TSplitter;
    MainVSplitter: TSplitter;
    View1: TMenuItem;
    miSettings: TMenuItem;
    miZoomBar: TMenuItem;
    redrawTimer: TTimer;
    ZoomImage: TImage;
    DevSectSize: TMenuItem;
    DS512: TMenuItem;
    DS2048: TMenuItem;
    DSCustom: TMenuItem;
    DS4096: TMenuItem;
    Extras1: TMenuItem;
    BlockInfo: TGroupBox;
    CKCol1Panel: TPanel;
    CKCol2Panel: TPanel;
    ShapeFinished: TShape;
    ShapeNonTried: TShape;
    ShapeBad: TShape;
    ShapeNonScraped: TShape;
    ShapeNonTrimmed: TShape;
    lblRescued: TLabel;
    lblNonTried: TLabel;
    lblBad: TLabel;
    lblNonScraped: TLabel;
    lblNonTrimmed: TLabel;
    ShapeActive: TShape;
    lblActive: TLabel;
    ZoomPanel: TPanel;
    miRefresh: TMenuItem;
    miAbout: TMenuItem;
    TopPanel: TPanel;
    TopPiePanel: TPanel;
    PieImage: TImage;
    MainMenu1: TMainMenu;
    FileMenu: TMenuItem;
    Openlogfile1: TMenuItem;
    Exit1: TMenuItem;
    BlockImage: TImage;
    OpenDialog1: TOpenDialog;
    StatusBar1: TStatusBar;
    DbgLog: TMemo;
    Closefile1: TMenuItem;
    OptionsMenu: TMenuItem;
    Automatc1: TMenuItem;
    N5sec1: TMenuItem;
    N10sec1: TMenuItem;
    N30sec1: TMenuItem;
    N1min1: TMenuItem;
    off1: TMenuItem;
    N2mins1: TMenuItem;
    N5mins1: TMenuItem;
    updateTimer: TTimer;
    GridPanel: TPanel;
    Prefixes1: TMenuItem;
    BinaryKiBMiB1: TMenuItem;
    DecimalKBMB1: TMenuItem;
    Gridsize1: TMenuItem;
    N4px1: TMenuItem;
    N6px1: TMenuItem;
    N8px1: TMenuItem;
    N10px1: TMenuItem;
    N12px1: TMenuItem;
    N14px1: TMenuItem;
    N16px1: TMenuItem;
    N20px1: TMenuItem;
    N24px1: TMenuItem;
    ApplicationEvents1: TApplicationProperties;
    procedure BlockImageMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure BtnCloseZoomBarClick(Sender: TObject);
    procedure miDevExampleClick(Sender: TObject);
    procedure miSettingsClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BlockImageMouseDown(Sender: TObject; {%H-}Button: TMouseButton; {%H-}Shift: TShiftState; X, Y: Integer);
    procedure DSClick(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure FormResize(Sender: TObject);
    procedure Closefile1Click(Sender: TObject);
    procedure miShowLogClick(Sender: TObject);
    procedure miZoomBarClick(Sender: TObject);
    procedure setGridSize(Sender: TObject);
    procedure Openlogfile1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure miRefreshClick(Sender: TObject);
    procedure autoParseClick(Sender: TObject);
    procedure redrawTimerTimer(Sender: TObject);
    procedure UnitFormatClick(Sender: TObject);
    procedure miAboutClick(Sender: TObject);
    procedure miClearlogClick(Sender: TObject);
    procedure ApplicationEvents1ShowHint(var HintStr: String;
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure ZoomImageMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ZoomImageMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ZoomImageMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    FParser : TLogParser;
    FGridBlockStore : TImgBlockStore;
    FZoomBlockStore : TImgBlockStore;
    Fhblocks : integer;
    Fvblocks : integer;
    FgridSize : integer;
    FDraggingZoom : boolean;
    FDraggingZoomOrigin : Longint;
    FSelectedBlock : Longint;
    procedure drawBlocks;
    procedure doDrawBlocks;
    procedure drawBlocksFast;
    procedure drawZoomImage;
    procedure updateStatusTexts;
    function getImgBlockFrameRect(imgBlock : integer):TRect;
    function getImgBlockRect(imgBlock : integer):TRect;
  public
    { Public declarations }
    // for the Observer role
    Procedure FPOObservedChanged(ASender : TObject;
      Operation : TFPObservedOperation; Data : Pointer);
    // for the Observed role
    procedure AttachObserver(AObserver : TObject);
    procedure DetachObserver(AObserver : TObject);
    procedure SelectBlock(block : integer);
    function GridXYToBlock(X, Y : integer) : integer;
    property Parser : TLogParser read FParser;
    property GridBlockStore : TImgBlockStore read FGridBlockStore;
    property ZoomBlockStore : TImgBlockStore read FZoomBlockStore;
  end;


var
  MainForm: TMainForm;

implementation
uses Math, BlockInspector, About, Settings;

{$R *.lfm}

procedure TMainForm.AttachObserver(AObserver : TObject);
begin
  FPOAttachObserver(AObserver);
end;

procedure TMainForm.DetachObserver(AObserver : TObject);
begin
  FPODetachObserver(AObserver);
end;

// initialization on program start
procedure TMainForm.FormCreate(Sender: TObject);
var i : integer;
begin
  DoubleBuffered:=true;
  TopPanel.DoubleBuffered:=true;
  MainPanel.DoubleBuffered:=true;
  StatusBar1.DoubleBuffered:=true;
  BottomPanel.DoubleBuffered:=true;

  FgridSize := 8;

  ZoomImage.Align:=alNone;  // alClient only useful for Form Designer
  BlockImage.Align:=alNone; // at x-platform designtime, but disruptive at runtime
  PieImage.Picture.Bitmap.SetSize(PieImage.Width, PieImage.Height);

  FParser := TLogParser.Create;

  FZoomBlockStore := TImgBlockStore.Create(self);
  FZoomBlockStore.DataSource := FParser; // must be before FGridBlockStore's
  FZoomBlockStore.AttachObserver(self);

  FGridBlockStore := TImgBlockStore.Create(self);
  FGridBlockStore.DataSource := FParser;
  FGridBlockStore.AttachObserver(self);

  FormResize(self); // to trigger initial block display
  Constraints.MinWidth:=RescueStatusBox.Left+RescueStatusBox.Width+64+4;

  // open first existing file specified in command line
  for i:=1 to ParamCount do if FileExists(ParamStr(i)) then begin
    FParser.OpenFile(ParamStr(i));
  end;
end;

// destruction when form is closed
procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FGridBlockStore.DetachObserver(self);
  FZoomBlockStore.DetachObserver(self);
  FreeAndNil(FGridBlockStore);
  FreeAndNil(FZoomBlockStore);
  FreeAndNil(FParser);
end;

// Mainform is notified about a change
procedure TMainForm.FPOObservedChanged(ASender: TObject;
  Operation: TFPObservedOperation; Data: Pointer);
begin
  if (ASender = FGridBlockStore) then begin
    if FParser.hasFile() then
    Caption := PROGRAM_TITLE + ' [' +
               Copy(FGridBlockStore.DataSource.LogFileName, LastDelimiter('\/:',
                    FGridBlockStore.DataSource.LogFileName)+1, 128) + ']'
    else Caption := PROGRAM_TITLE;
    DSClick(self);
    drawBlocks;
    updateStatusTexts;
  end;
  if (ASender = FZoomBlockStore) then drawZoomImage;
end;

// file(s) have been dropped onto the form. Open the first file.
procedure TMainForm.FormDropFiles(Sender: TObject; const FileNames: array of String);
begin
  FParser.OpenFile(FileNames[0]);
end;

// actually called now when the MainPanel resizes. It contains the controls
// that need manual care on resize.
procedure TMainForm.FormResize(Sender: TObject);
const GridPanelBorder = 4;
var
  hBlocksBefore, vBlocksBefore : integer;
  nBlocksChanged, zoomImageChanged: Boolean;
begin
  (* BlockImage resizing *)
  hBlocksBefore := FhBlocks;
  vBlocksBefore := FvBlocks;
  BlockImage.Width:=GridPanel.Width - GridPanelBorder - ((GridPanel.Width - GridPanelBorder-1) mod FgridSize);
  BlockImage.Height:=GridPanel.Height - GridPanelBorder - ((GridPanel.Height - GridPanelBorder-1) mod FgridSize);
  FhBlocks := BlockImage.Width div FgridSize;  // number of horizontal blocks in display
  FvBlocks := BlockImage.Height div FgridSize; // same for vertical

  nBlocksChanged := (hBlocksBefore <> Fhblocks) or (vBlocksBefore <> FvBlocks);
  if nBlocksChanged then begin
    BlockImage.Picture.Bitmap.SetSize(BlockImage.Width, BlockImage.Height);
    FGridBlockStore.numBlocks:=FhBlocks*FvBlocks;
  end;

  (* ZoomImage resizing: Ensure same height as BlockImage *)
  zoomImageChanged := ZoomImage.Height<>BlockImage.Height;
  if zoomImageChanged then begin
    ZoomImage.Height:=BlockImage.Height;
    ZoomImage.Width:=ZoomPanel.Width-GridPanelBorder;
    ZoomImage.Picture.Bitmap.SetSize(ZoomImage.Width, ZoomImage.Height);
    FZoomBlockStore.numBlocks:=ZoomImage.Height;
  end;
  if zoomImageChanged or nBlocksChanged then begin
    SelectBlock(-1); // unselect block
    FPONotifyObservers(self, ooChange, nil); // trigger imgblock updates
  end;
end;

// update a lot of textual info on the form.
procedure TMainForm.updateStatusTexts;
begin
  with FParser do begin
    if hasFile() then begin
      CommentsMemo.Clear;
      CommentsMemo.Lines.AddStrings(CommentLines);

      RescueStatusBox.Caption:='Rescue status:  '+OperationToText(rescueStatus.curOperation);
      EditInputSize.Text:=SizeStr(rescueStatus.devicesize);
      EditWhatever.Text:='here';
      EditCurrentPos.Text:=SizeStr(rescueStatus.pos);
      EditErrorSize.Text:=SizeStr(rescueStatus.bad + rescueStatus.nonscraped +
                                  rescueStatus.nontrimmed);
      EditRescued.Text:=SizeStr(rescueStatus.rescued);
      EditErrorCount.Text:=IntToStr(rescueStatus.errors);
      StatusBar1.Panels.Items[0].Text:=LogFileName;
    end else begin
      RescueStatusBox.Caption:='Rescue status';
      EditInputSize.Text:=SizeStr(0);
      EditWhatever.Text:='here';
      EditCurrentPos.Text:=SizeStr(0);
      EditErrorSize.Text:=SizeStr(0);
      EditRescued.Text:=SizeStr(0);
      EditErrorCount.Text:='0';
      StatusBar1.Panels.Items[0].Text:='No file opened';
    end;
  end; // with FParser
end;

// constructs a TRect containing the requested Image Block
function TMainForm.getImgBlockRect(imgBlock : integer):TRect;
var hblock, vblock: integer;
begin
   vblock := imgBlock div Fhblocks;
   hblock := imgBlock mod Fhblocks;
   result.Left:=1+hblock*FgridSize;
   result.Top:=1+vblock*FgridSize;
   result.Right:=result.Left+FgridSize-1;
   result.Bottom:=result.Top+FgridSize-1;
end;

// constructs a TRect containing the requested Image Block, including border
function TMainForm.getImgBlockFrameRect(imgBlock : integer):TRect;
begin
   result:=getImgBlockRect(imgBlock);
   result.Left:=result.Left-1;
   result.Top:=result.Top-1;
   result.Right:=result.Right+1;
   result.Bottom:=result.Bottom+1;
end;

procedure TMainForm.drawBlocks;
begin
  redrawTimer.Enabled:=true;
end;

//get block colors from GridBlockStore and draw them on the BlockImage
procedure TMainForm.doDrawBlocks;
var iImgBlock : Longint;
    curImgBlock : Longword;
    blockRect : TRect;
    angle, alength: extended;
begin
  with BlockImage.Canvas do begin
    Brush.Color := COLOR_UNDEFINED; // color for the gridlines
    FillRect(ClipRect);
    for iImgBlock := 0 to FGridBlockStore.numBlocks-1 do begin
      curImgBlock := FGridBlockStore[iImgBlock]; // get current block
      Brush.Color:=curImgBlock and $ffffff; //extract color
      if LongBool(curImgBlock and MASK_ACTIVE) then begin // current pos marker
        blockRect:=getImgBlockFrameRect(iImgBlock);
        Pen.Color := COLOR_ACTIVE;
        Rectangle(blockRect);
      end else begin
        blockRect:=getImgBlockRect(iImgBlock);
        FillRect(blockRect);
      end;
    end;
    if FSelectedBlock >= 0 then begin
      blockRect:=getImgBlockFrameRect(FSelectedBlock);
      blockRect.Top-=1; blockRect.Left-=1;
      blockRect.Bottom+=1; blockRect.Right+=1;
      Brush.Color := clWhite;
      FrameRect(blockRect);
    end;
  end;


  //DbgLog.Lines.Add('drawn image blocks: '+inttostr(FGridBlockStore.numBlocks));

  //Example blocks in the top bar
  ShapeFinished.Brush.Color:=COLOR_FINISHED;
  ShapeNonTried.Brush.Color:=COLOR_NON_TRIED;
  ShapeBad.Brush.Color:=COLOR_BAD_SECT;
  ShapeNonScraped.Brush.Color:=COLOR_NON_SCRAPED;
  ShapeNonTrimmed.Brush.Color:=COLOR_NON_TRIMMED;
  ShapeActive.Pen.Color:=COLOR_ACTIVE;

  // Build Pie image
  PieImage.Canvas.Brush.Color := TopPiePanel.Color;
  PieImage.Canvas.FillRect(PieImage.Canvas.ClipRect);
  with FParser do begin
    if hasFile() then begin
      PieImage.Canvas.Brush.Color := COLOR_FINISHED;
      alength:=5760*rescueStatus.rescued/rescueStatus.devicesize;
      PieImage.Canvas.RadialPie(1, 1, 63, 63, 0, round(alength));
      angle:=alength;

      PieImage.Canvas.Brush.Color := COLOR_NON_TRIED;
      alength:=5760*rescueStatus.nontried/rescueStatus.devicesize;
      PieImage.Canvas.RadialPie(1, 1, 63, 63, round(angle), round(alength));
      angle:=angle+alength;

      PieImage.Canvas.Brush.Color := COLOR_BAD_SECT;
      alength:=5760*rescueStatus.bad/rescueStatus.devicesize;
      PieImage.Canvas.RadialPie(1, 1, 63, 63, round(angle), round(alength));
      angle:=angle+alength;

      PieImage.Canvas.Brush.Color := COLOR_NON_SCRAPED;
      alength:=5760*rescueStatus.nonscraped/rescueStatus.devicesize;
      PieImage.Canvas.RadialPie(1, 1, 63, 63, round(angle), round(alength));
      angle:=angle+alength;

      PieImage.Canvas.Brush.Color := COLOR_NON_TRIMMED;
      alength:=5760*rescueStatus.nontrimmed/rescueStatus.devicesize;
      PieImage.Canvas.RadialPie(1, 1, 63, 63, round(angle), round(alength));
    end;
  end;
end;

// test with direct pixel access. Currently unused, seems not so
// much faster. most time is spent updating the image on screen after drawing.
procedure TMainForm.drawBlocksFast;
var
    rowSize, x, y, imgBlock, yBlocks, numBlocks : Longint;
    PixPtr, RowPtr: PColor;
    RawImage: TRawImage;
    BytePerLine: Integer;
    isUndef, ygridline : LongBool;
begin
  try
    BlockImage.Picture.Bitmap.BeginUpdate(false);
    RawImage := BlockImage.Picture.Bitmap.RawImage;
    RowPtr := PColor(RawImage.Data);
    BytePerLine := RawImage.Description.BytesPerLine;
    numBlocks := FGridBlockStore.numBlocks;
    for y:=0 to RawImage.Description.Height-1 do begin
      PixPtr := RowPtr; //point to start of row
      yBlocks := (y div FgridSize)*Fhblocks;
      ygridline := (y mod Fgridsize)=0;
      for x:=0 to RawImage.Description.Width-1 do begin    // 1-pixel steps
        imgBlock:=yBlocks + x div FgridSize;
        isUndef:= ygridline or ((x mod Fgridsize)=0) or (imgBlock>=numBlocks);
        imgBlock:=imgBlock mod numBlocks;
        PixPtr^ := $ffffff and (COLOR_UNDEFINED*Longint(isUndef) + FGridBlockStore[imgBlock]*Longint(not isUndef));
        Inc(PixPtr, 1); // advance 1 pixel
      end;
      Inc(PByte(RowPtr), BytePerLine); // get next row start
    end;
  finally
    BlockImage.Picture.Bitmap.EndUpdate(false);
  end;

end;

// the user has scrolled the mouse wheel on the block grid -> perform zoom
procedure TMainForm.BlockImageMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var
  block : Longint; // wheeled-on imgblock
begin
  SelectBlock(-1); // unselect block
  block:=GridXYToBlock(MousePos.x, MousePos.y);
  if WheelDelta>0 then begin // zoom in
    FGridBlockStore.zoom(block, 1/1.5);
  end else if WheelDelta<0 then begin // zoom out
    FGridBlockStore.zoom(block, 1.5);
  end;
  FPONotifyObservers(self, ooChange, nil); // get update
end;

// draw the left-hand side zooming bar using the blocks in FZoomBlockStore
procedure TMainForm.drawZoomImage;
var
  iImgBlock, zoomWndFirstPx, zoomWndLastPx, cPosBlock : Longint;
  curImgBlock : Longword;
begin
  cPosBlock:=-3;
  with ZoomImage.Canvas do begin
    // Draw the block colors
    Line(0, 0, Width, 0); // prevents a canvas glitch
    for iImgBlock := 0 to FZoomBlockStore.numBlocks-1 do begin
      curImgBlock := FZoomBlockStore[iImgBlock]; // get current block
      Pen.Color:=curImgBlock and $ffffff; //extract color
      Line(0, iImgBlock, Width, iImgBlock);
      if LongBool(curImgBlock and MASK_ACTIVE) then cPosBlock:=iImgBlock // current pos marker
    end;
    // Draw the current position marker
    Pen.Color := COLOR_ACTIVE;
    Line(0,cPosBlock-1, 3, cPosBlock-1);
    Line(Width-3,cPosBlock-1, Width, cPosBlock-1);
    Line(0,cPosBlock, 5, cPosBlock);
    Line(Width-5,cPosBlock, Width, cPosBlock);
    Line(0,cPosBlock+1, 3, cPosBlock+1);
    Line(Width-3,cPosBlock+1, Width, cPosBlock+1);
    // Draw the zoom frame
    if FZoomBlockStore.DataSource.hasFile() then begin
      Pen.Mode:=pmMask;
      Pen.Color:=clBlack;
      Pen.Style:=psDot;
      zoomWndFirstPx:=FZoomBlockStore.getBlockFromOffset(FGridBlockStore.devBSize*
        FGridBlockStore.BlockOffset);
      zoomWndLastPx:=FZoomBlockStore.getBlockFromOffset(FGridBlockStore.devBSize*
        (FGridBlockStore.BlockOffset+FGridBlockStore.BlockCount));
      Line(0, zoomWndFirstPx, Width, zoomWndFirstPx);  // top border
      Line(0, zoomWndFirstPx, 0, zoomWndLastPx);       // left border
      Line(Width-1, zoomWndFirstPx, Width-1, zoomWndLastPx); // right border
      Line(0, zoomWndLastPx, Width, zoomWndLastPx);          // bottom border
      Pen.Style:=psSolid;
      Pen.Mode:=pmCopy;
    end;
  end;
end;

// select an img block from the block grid. This happens when a bock is clicked.
procedure TMainForm.SelectBlock(block : integer);
begin
  FSelectedBlock:=block;
  if Assigned(BlockForm) then BlockForm.Block:=block;
  drawBlocks;
end;


// user wants to open a file (file open dialog)
procedure TMainForm.Openlogfile1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then FParser.OpenFile(OpenDialog1.FileName);
end;

// user wants to close current file
procedure TMainForm.Closefile1Click(Sender: TObject);
begin
  BlockForm.Hide;
  FParser.CloseFile;
end;

// user wants to show the bottom panel (debug log and logfile comments)
procedure TMainForm.miShowLogClick(Sender: TObject);
begin
  BottomPanel.Visible:=not BottomPanel.Visible;
  miShowLog.Checked:=BottomPanel.Visible;
  MainVSplitter.Visible:=BottomPanel.Visible;
  FormResize(self);
end;


// user wants to toggle the zoom bar
procedure TMainForm.miZoomBarClick(Sender: TObject);
begin
  ZoomPanel.Visible:=not ZoomPanel.Visible;
  miZoomBar.Checked:=ZoomPanel.Visible;
  FormResize(self);
end;


// user clicked exit from the menu
procedure TMainForm.Exit1Click(Sender: TObject);
begin
  BlockForm.Hide;
  SettingsForm.Hide;
  Application.Terminate;
end;

// refresh requested, either from the menu or programmatically
procedure TMainForm.miRefreshClick(Sender: TObject);
begin
  FParser.parse();
end;

procedure TMainForm.BtnCloseZoomBarClick(Sender: TObject);
begin
  miZoomBarClick(self);
end;

procedure TMainForm.miDevExampleClick(Sender: TObject);
begin
  FParser.OpenFile('..\example_logs\simon.log');
end;

procedure TMainForm.miSettingsClick(Sender: TObject);
begin
  SettingsForm.Show;
end;

// maps a X,Y pixel location on the block image to a block number
function TMainForm.GridXYToBlock(X, Y : integer) : integer;
begin
  result := (Y div FgridSize)*Fhblocks + X div FgridSize;
end;

// user clicked the Block Grid
procedure TMainForm.BlockImageMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var imgBlock: integer;
begin
  if (Button<>mbLeft) then exit;
  if (X mod FgridSize = 0) or (Y mod FgridSize = 0) then exit; // user clicked gridlines
  if not FParser.hasFile() then exit;  // empty log, no file opened
  imgBlock := GridXYToBlock(X, Y); // calculate clicked block
  SelectBlock(imgBlock);

  // Block Inspector initialization
  if Assigned(BlockForm) then begin
    if not BlockForm.Visible then BlockForm.Show;
  end;
end;

// selected a device sector size from the menu, by user or programmatically
procedure TMainForm.DSClick(Sender: TObject);
var strSize : string;
    itemID : integer;
begin
  DSCustom.Checked:=true; // initially select the custom entry

  // determine the size from the clicked item
  if Sender is TMenuItem then begin
    TMenuItem(Sender).Checked:=true;
    if Sender = DSCustom then begin
      strSize := InputBox('Custom sector size', 'Enter custom sector size in Bytes:',
        inttostr(FGridBlockStore.devBSize));
      FGridBlockStore.devBSize := strToIntDef(strSize, DEF_BSIZE);
    end else FGridBlockStore.devBSize:=TMenuItem(Sender).Tag; // size is stored in .Tag
    FZoomBlockStore.devBSize:=FGridBlockStore.devBSize; //adjust zoomBlockStore accordingly
  end;


  // depending on the size, determine which item has to be selected
  for itemID := 0 to DevSectSize.Count-1 do begin
    if DevSectSize.Items[itemID].Tag = FGridBlockStore.devBSize then begin
      DevSectSize.Items[itemID].Checked:=true;
      break;
    end;
  end;
  if DSCustom.Checked then DSCustom.Caption :=
    'Custom ('+inttostr(FGridBlockStore.devBSize) + ') ...'
    else DSCustom.Caption := 'Custom...';

  // trigger update (notify gridblockstore)
  if Sender is TMenuItem then FPONotifyObservers(self, ooChange, nil);
end;

// user selected an automatic parse time interval from the menu
procedure TMainForm.autoParseClick(Sender: TObject);
var interval : integer;
begin
  if Sender is TMenuItem then begin
    TMenuItem(Sender).Checked := true;
    interval := TMenuItem(Sender).Tag * 1000;
    updateTimer.Enabled := interval <> 0;
    updateTimer.Interval := interval;
  end;
  FParser.parse();
end;

procedure TMainForm.redrawTimerTimer(Sender: TObject);
begin
  redrawTimer.Enabled:=false;
  doDrawblocks;
end;


// user selected a unit system from the menu
procedure TMainForm.UnitFormatClick(Sender: TObject);
begin
  if Sender is TMenuItem then TMenuItem(Sender).Checked := true;
  useDecimalUnits := DecimalKBMB1.Checked;
  FPONotifyObservers(self, ooChange, nil); // also push change to other forms
end;

// user selected a grid size from the menu
procedure TMainForm.setGridSize(Sender: TObject);
begin
  if Sender is TMenuItem then begin
    TMenuItem(Sender).Checked := true;
    FgridSize:=TMenuItem(Sender).Tag; // size stored in tag field
  end;
  FormResize(self);
end;

// shows the about box
procedure TMainForm.miAboutClick(Sender: TObject);
begin
  AboutBox.Version.Caption := 'Version ' + VERSION_MAJOR + '.' + VERSION_MINOR;
  AboutBox.ShowModal;
  self.Show; // workaround: avoid mainform from losing focus after aboutbox closes
end;

// clears the log
procedure TMainForm.miClearlogClick(Sender: TObject);
begin
  DbgLog.Lines.Clear;
end;

// modify mouse-over Hint time on a string length basis
procedure TMainForm.ApplicationEvents1ShowHint(var HintStr: String; var CanShow: Boolean; var HintInfo: THintInfo);
begin
  HintInfo.HideTimeout:=2000+Length(HintStr)*100;
end;

procedure TMainForm.ZoomImageMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if FZoomBlockStore.DataSource.hasFile() then begin
    if ZoomImage.Cursor=crHandPoint then begin
      FDraggingZoom:=true;
      FDraggingZoomOrigin:=Y;
    end;
  end;
end;

procedure TMainForm.ZoomImageMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  zoomWndFirstPx, zoomWndLastPx : Longint;
  zoomShift, deviceSectors : Int64;  // all values in device sectors
begin
  if FZoomBlockStore.DataSource.hasFile() then begin
    if FDraggingZoom and (FDraggingZoomOrigin<>Y) then begin // user dragged zoom area
      zoomShift:=(Y-FDraggingZoomOrigin)*FZoomBlockStore.bytesPerBlock div FGridBlockStore.devBSize;
      deviceSectors:=FGridBlockStore.DataSource.rescueStatus.devicesize div FGridBlockStore.devBSize;
      FGridBlockStore.BlockOffset:=min(deviceSectors-FGridBlockStore.BlockCount,
        max(0, FGridBlockStore.BlockOffset+zoomShift));
      FDraggingZoomOrigin:=Y;
      FPONotifyObservers(self, ooChange, nil);
    end;
    zoomWndFirstPx:=FZoomBlockStore.getBlockFromOffset(FGridBlockStore.devBSize*FGridBlockStore.BlockOffset);
    zoomWndLastPx:=FZoomBlockStore.getBlockFromOffset(FGridBlockStore.devBSize*
      (FGridBlockStore.BlockOffset+FGridBlockStore.BlockCount));
    if (Y>=zoomWndFirstPx) and (Y<=zoomWndLastPx) then ZoomImage.Cursor:=crHandPoint
      else ZoomImage.Cursor:=crDefault;
  end;
end;

procedure TMainForm.ZoomImageMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FDraggingZoom:=false;
end;

end.

